// Copyright (c) 2015, Robert Escriva
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of Replicant nor the names of its contributors may be
//       used to endorse or promote products derived from this software without
//       specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#define __STDC_LIMIT_MACROS

// e
#include <e/strescape.h>

// Replicant
#include <replicant.h>
#include "tools/common.h"

int
main(int argc, const char* argv[])
{
    const char* obj = "replicant";
    const char* cond = "configuration";
    long start_state = 0;
    bool follow = false;
    connect_opts conn;
    e::argparser ap;
    ap.autohelp();
    ap.option_string("[OPTIONS]");
    ap.arg().name('o', "object")
            .description("object that maintains the condition")
            .metavar("OBJ").as_string(&obj);
    ap.arg().name('c', "cond")
            .description("condition name")
            .metavar("COND").as_string(&cond);
    ap.arg().name('i', "initial-state")
            .description("initial state of the condition")
            .metavar("S").as_long(&start_state);
    ap.arg().long_name("follow")
            .description("use cond_follow instead of repeated cond_wait")
            .set_true(&follow);
    ap.add("Connect to a cluster:", conn.parser());

    if (!ap.parse(argc, argv))
    {
        return EXIT_FAILURE;
    }

    if (ap.args_sz() != 0)
    {
        std::cerr << "command takes no positional arguments\n" << std::endl;
        ap.usage();
        return EXIT_FAILURE;
    }

    if (!conn.validate())
    {
        std::cerr << "invalid host:port specification\n" << std::endl;
        ap.usage();
        return EXIT_FAILURE;
    }

    try
    {
        replicant_client* r = replicant_client_create(conn.host(), conn.port());

        if (follow)
        {
            uint64_t state = 0;
            char* data = NULL;
            size_t data_sz = 0;
            replicant_returncode re = REPLICANT_GARBAGE;
            int64_t rid = replicant_client_cond_follow(r, obj, cond, &re, &state, &data, &data_sz);

            while (true)
            {
                if (!cli_finish(r, rid, &re))
                {
                    return EXIT_FAILURE;
                }

                std::cout << "condition >= " << state
                          << " \"" << e::strescape(std::string(data, data_sz)) << "\""
                          << std::endl;
            }
        }
        else
        {
            for (uint64_t x = start_state; ; ++x)
            {
                char* data = NULL;
                size_t data_sz = 0;
                replicant_returncode re = REPLICANT_GARBAGE;
                int64_t rid = replicant_client_cond_wait(r, obj, cond, x, &re, &data, &data_sz);

                if (!cli_finish(r, rid, &re))
                {
                    return EXIT_FAILURE;
                }

                std::cout << "condition >= " << x
                          << " \"" << e::strescape(std::string(data, data_sz)) << "\""
                          << std::endl;
            }
        }

        return EXIT_SUCCESS;
    }
    catch (std::exception& e)
    {
        std::cerr << "error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
}
